import boto3
from typing import List, Dict, Any, Optional
from flotorch_core.logger.global_logger import get_logger
from flotorch_core.storage.db.vector.vector_storage import VectorStorage, VectorStorageSearchItem, VectorStorageSearchResponse
from flotorch_core.embedding.embedding import BaseEmbedding, EmbeddingMetadata


logger = get_logger()


class BedrockKnowledgeBaseStorage(VectorStorage):
    def __init__(self, knowledge_base_id: str, region: str = 'us-east-1', embedder: Optional[BaseEmbedding] = None):
        self.client = boto3.client("bedrock-agent-runtime", region_name=region)
        self.knowledge_base_id = knowledge_base_id

    def search(self, chunk, knn: int, hierarchical: bool = False) -> VectorStorageSearchResponse:
        """
        Searches the Bedrock Knowledge Base using vector search.
        """
        query = {"text": chunk.data}
        retrieval_configuration = {
            'vectorSearchConfiguration': {'numberOfResults': knn}
        }

        try:
            response = self.client.retrieve(
                knowledgeBaseId=self.knowledge_base_id,
                retrievalQuery=query,
                retrievalConfiguration=retrieval_configuration
            )
            formatted_results = self._format_response(response)

            return VectorStorageSearchResponse(
                status=True,
                result=formatted_results,
                metadata={
                    "embedding_metadata": EmbeddingMetadata(0, 0)
                }
            )

        except Exception as e:
            logger.error(f"Error retrieving from Bedrock Knowledge Base: {str(e)}")
            return VectorStorageSearchResponse(
                status=False,
                result=[],
                metadata={"error": str(e)}
            )

    def _format_response(self, data) -> List[VectorStorageSearchItem]:
        formatted_results = []
        for result in data.get('retrievalResults', []):
            content = result.get('content', {})
            text = content.get('text', '')

            if text:
                formatted_results.append(VectorStorageSearchItem(text=text))

        return formatted_results

    def embed_query(self, query_vector: List[float], knn: int, hierarchical: bool = False) -> Dict[str, Any]:
        """
        Bedrock Knowledge Base does not support explicit query vectorization, as embedding is managed internally.
        """
        raise NotImplementedError("Embedding is managed internally by Bedrock Knowledge Base.")

    def read(self, key) -> dict:
        raise NotImplementedError("Not implemented")

    def write(self, item: dict):
        raise NotImplementedError("Not implemented.")


import logging
import os
from typing import Generator
from urllib.parse import urlparse
import boto3
from .storage import StorageProvider

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class S3StorageProvider(StorageProvider):
    """
    S3 storage provider
    """

    def __init__(self, bucket: str, s3_client = boto3.client('s3')):

        """
        Initializes the S3Storage class with the specified S3 bucket.
        Args:
            bucket (str): The name of the S3 bucket to interact with.
        Attributes:
            bucket (str): The name of the S3 bucket.
            s3_client (boto3.client): The boto3 client for interacting with S3.
        """

        super().__init__()
        self.bucket = bucket
        self.s3_client = s3_client

    def get_path(self, uri: str) -> str:
        parsed = urlparse(uri)
        return parsed.path.lstrip("/")

    def write(self, path: str, data: bytes) -> None:
        """
        Writes data to the specified path in the S3 bucket.
        Args:
            path (str): The path to write the data to in the S3 bucket.
            data (bytes): The data to write to the S3 bucket.
        """
        logger.info(f'Writing data to S3 storage: {data}')
        if not path.endswith("/"):
            key = path
        else:
            key = path + "tmp.data"
        self.s3_client.put_object(Bucket=self.bucket, Key=key, Body=data)

    def read(self, path: str) -> Generator[bytes, None, None]:
        """
        Reads data from the specified path in the S3 bucket.
        Args:
            path (str): The path to read the data from in the S3 bucket.
        Returns:
            Generator[bytes, None, None]: A generator that yields the data read from the S3 bucket.
        """
        logger.info('Reading data from S3 storage')
        if self._is_directory(path):
            yield from self._read_directory(path)
        else:
            response = self.s3_client.get_object(Bucket=self.bucket, Key=path)
            yield response['Body'].read()

    def _is_directory(self, path: str) -> bool:
        """
        Determines if the given S3 path is a directory by checking if multiple files exist under the prefix.

        Args:
            path (str): The S3 path.

        Returns:
            bool: True if the path is a directory, False otherwise.
        """
        if not path.endswith("/"):  # Ensure the path ends with a slash to treat it as a directory
            path += "/"

        response = self.s3_client.list_objects_v2(Bucket=self.bucket, Prefix=path, MaxKeys=1)
        return "Contents" in response  # If "Contents" exists, it's a directory

    def _read_directory(self, path: str) -> Generator[bytes, None, None]:
        """
        Reads all files in the specified S3 directory.

        Args:
            path (str): The S3 directory path.

        Returns:
            Generator[bytes, None, None]: Yields file contents.
        """
        if not path.endswith("/"):
            path += "/"

        response = self.s3_client.list_objects_v2(Bucket=self.bucket, Prefix=path)

        if "Contents" not in response:
            logger.warning(f"No files found in S3 directory: {path}")
            return

        for obj in response["Contents"]:
            file_key = obj["Key"]
            if file_key.endswith("/"):  # Skip directories
                continue

            logger.info(f"Reading S3 file: {file_key}")
            file_response = self.s3_client.get_object(Bucket=self.bucket, Key=file_key)
            yield file_response['Body'].read()


from typing import List

from langchain.text_splitter import CharacterTextSplitter

from .chunking import BaseChunker, Chunk


class FixedSizeChunker(BaseChunker):
    """
    This class is responsible for chunking the text into fixed size chunks.
    :param chunk_size: The size of the chunk.
    :param chunk_overlap: The overlap between chunks.
    """
    def __init__(self, chunk_size: int, chunk_overlap: int):
        super().__init__()
        self.chunk_size = self.tokens_per_charecter * chunk_size
        self.chunk_overlap = int(chunk_overlap * self.chunk_size / 100)
        if self.chunk_size <= 0:
            raise ValueError("chunk_size must be positive")
        if self.chunk_overlap >= self.chunk_size:
            raise ValueError("chunk_overlap must be less than chunk_size")

    """
    Chunks the text into fixed size chunks.
    :param data: The input text.
    :return: The list of chunks."""
    def chunk(self, data: str) -> List[Chunk]:
        if not data:
            raise ValueError("Input text cannot be empty or None")

        data = self._clean_data(data)
        self.text_splitter = CharacterTextSplitter(
            separator=self.space,
            chunk_size=self.chunk_size,
            chunk_overlap=self.chunk_overlap,
            length_function=len,
            is_separator_regex=False
        )
        chunks = self.text_splitter.split_text(data)
        return [Chunk(chunk) for chunk in chunks]



from typing import List

import ollama
from .embedding import BaseEmbedding, Embeddings, EmbeddingMetadata
from flotorch_core.chunking.chunking import Chunk
from .embedding_registry import register

"""

This class is responsible for embedding the text using the Llama model.
If Ollama server is running remotely set the environment variable OLLAMA_HOST to the server URL."""
@register("llama2")
class LlamaEmbedding(BaseEmbedding):
    """
    Initializes the LlamaEmbedding class.
    :param model_id: The model id of the Llama model.
    """

    def __init__(self, model_id: str, region: str, dimensions: int = 256, normalize: bool = True):
        super().__init__(model_id, region, dimensions, normalize)

    """
    Prepares the chunk for embedding.
    :param chunk: The chunk to be embedded.
    :return: The prepared chunk.
    """
    def embed(self, chunk: Chunk) -> Embeddings:
        response = ollama.embeddings(model=self.model_id, prompt=chunk.data)
        embedding = Embeddings(embeddings=response['embedding'], metadata=None)
        return embedding


import unittest

from chunking.chunking import Chunk
from chunking.fixedsize_chunking import FixedSizeChunker
from chunking.hierarical_chunking import HieraricalChunker


class TestChunk(unittest.TestCase):
    def test_chunk_initialization(self):
        data = "Sample data"
        chunk = Chunk(data)
        self.assertEqual(chunk.data, data)
        self.assertIsNone(chunk.child_data)
        self.assertTrue(chunk.id)

    def test_add_child(self):
        chunk = Chunk("Parent data")
        chunk.add_child("Child data 1")
        chunk.add_child("Child data 2")
        self.assertEqual(chunk.child_data, ["Child data 1", "Child data 2"])

    def test_chunk_str(self):
        chunk = Chunk("Parent data")
        chunk.add_child("Child data 1")
        chunk.add_child("Child data 2")
        self.assertEqual(str(chunk), "Parent Chunk: Parent data, Chunk: ['Child data 1', 'Child data 2']")


class TestFixedSizeChunk(unittest.TestCase):
    def setUp(self):
        self.chunk_size = 10
        self.chunk_overlap = 20  # This should be less than chunk_size, change this value
        self.fixed_chunk = FixedSizeChunker(self.chunk_size, self.chunk_overlap)

    def test_initialization(self):
        # Assert chunking size and overlap are calculated correctly
        self.assertEqual(self.fixed_chunk.chunk_size, 40)  # 4 tokens per character * chunk_size
        self.assertEqual(self.fixed_chunk.chunk_overlap, 8)  # 20% overlap

    def test_invalid_chunk_size(self):
        with self.assertRaises(ValueError):
            FixedSizeChunker(0, 20)

    def test_invalid_chunk_overlap(self):
        # Make sure chunk_overlap is smaller than chunk_size
        with self.assertRaises(ValueError):
            FixedSizeChunker(10, 100)  # chunk_overlap exceeds chunk_size

    def test_chunk_empty_data(self):
        with self.assertRaises(ValueError):
            list(self.fixed_chunk.chunk(""))  # Empty data should raise ValueError

    def test_chunk_functionality(self):
        data = "This is a test string for chunking."
        chunks = self.fixed_chunk.chunk(data)
        self.assertIsInstance(chunks, list)
        self.assertGreater(len(chunks), 0)
        for chunk in chunks:
            self.assertIsInstance(chunk, Chunk)


class TestHieraricalChunk(unittest.TestCase):
    def setUp(self):
        self.chunk_size = 10
        self.chunk_overlap = 20  # This should be less than chunk_size, change this value
        self.parent_chunk_size = 50
        self.hierarchical_chunk = HieraricalChunker(self.chunk_size, self.chunk_overlap, self.parent_chunk_size)

    def test_initialization(self):
        # Assert chunking size and overlap are calculated correctly
        self.assertEqual(self.hierarchical_chunk.chunk_size, 40)  # 4 tokens per character * chunk_size
        self.assertEqual(self.hierarchical_chunk.parent_chunk_size, 200)  # 4 tokens per character * parent_chunk_size

    def test_invalid_parent_chunk_size(self):
        with self.assertRaises(ValueError):
            HieraricalChunker(10, 20, 0)

        with self.assertRaises(ValueError):
            HieraricalChunker(10, 20, 5)  # Parent chunking size must be larger than child chunking size

    def test_chunk_empty_data(self):
        with self.assertRaises(ValueError):
            list(self.hierarchical_chunk.chunk(""))  # Empty data should raise ValueError

    def test_chunk_functionality(self):
        data = "This is a test string for hierarchical chunking. It should split into parent and child chunks."
        chunks = self.hierarchical_chunk.chunk(data)
        self.assertIsInstance(chunks, list)
        self.assertGreater(len(chunks), 0)
        for chunk in chunks:
            self.assertIsInstance(chunk, Chunk)
            self.assertIsNotNone(chunk.child_data)
            self.assertGreater(len(chunk.child_data), 0)

    def test_parent_child_relation(self):
        data = "This is a test string for hierarchical chunking. It should split into parent and child chunks."
        chunks = self.hierarchical_chunk.chunk(data)
        parent_chunk = chunks[0]
        self.assertIsInstance(parent_chunk, Chunk)
        self.assertIsInstance(parent_chunk.child_data, list)
        self.assertGreater(len(parent_chunk.child_data), 0)


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod
import time
from pydantic import BaseModel
import botocore
from flotorch_core.logger.global_logger import get_logger

logger = get_logger()

class RetryParams(BaseModel):
    max_retries: int
    retry_delay: int
    backoff_factor: int


class BotoRetryHandler(ABC):
    """Abstract class for retry handler"""
    
    @property
    @abstractmethod
    def retry_params(self) -> RetryParams:
        pass
    
    @property
    @abstractmethod
    def retryable_errors(self) -> set[str]:
        pass
        
        
    def __call__(self, func):
        def wrapper(*args, **kwargs):
            retries = 0
            retry_params = self.retry_params
            while retries < retry_params.max_retries:
                try:
                    return func(*args, **kwargs)
                except botocore.exceptions.ClientError as e:
                    error_code = e.response['Error']['Code']
                    if error_code in self.retryable_errors:
                        retries += 1
                        logger.error(f"Rate limit error in Bedrock converse (Attempt {retries}/{retry_params.max_retries}): {str(e)}")
                        
                        if retries >= retry_params.max_retries:
                            logger.error("Max retries reached. Could not complete Bedrock converse operation.")
                            raise
                        
                        backoff_time = retry_params.retry_delay * (retry_params.backoff_factor ** (retries - 1))
                        logger.info(f"Retrying in {backoff_time} seconds...")
                        time.sleep(backoff_time)
                    else:
                        # If it's not a rate limit error, raise immediately
                        raise
                except Exception as e:
                    # For any other exception, log and raise immediately
                    logger.error(f"Unexpected error in Bedrock converse: {str(e)}")
                    raise
            
        return wrapper


from typing import Any, Dict, List

from flotorch_core.chunking.chunking import Chunk
from .bedrock_embedding import BedRockEmbedding
from .embedding_registry import register

"""
This class is responsible for embedding the text using the Cohere model.
"""
@register("cohere.embed-multilingual-v3")
@register("cohere.embed-english-v3")
class CohereEmbedding(BedRockEmbedding):
    """
    Initializes the CohereEmbedding class.
    :param model_id: The model id of the Cohere model.
    :param region: The region of the Cohere model.
    :param dimensions: The dimensions of the embedding.
    :param normalize: Normalize the embedding.
    """

    def __init__(self, model_id: str, region: str, dimensions: int = 256, normalize: bool = True) -> None:
        super().__init__(model_id, region, dimensions, normalize)

    """
    Prepares the chunk for embedding.
    :param chunk: The chunk to be embedded.
    :return: The prepared chunk.
    """
    
    def _prepare_chunk(self, chunk: Chunk) -> Dict:
        return {"texts": [chunk.data], "input_type": "search_document"}
    
    def extract_embedding(self, response: Dict[str, Any]) -> List[float]:
        return response["embeddings"][0]




from flotorch_core.storage.db.vector.no_ops_vector_storage import NoOpsVectorStorage
from flotorch_core.storage.db.vector.open_search import OpenSearchClient
from flotorch_core.storage.db.vector.bedrock_knowledgebase_storage import BedrockKnowledgeBaseStorage
from flotorch_core.storage.db.vector.vector_storage import VectorStorage
from flotorch_core.embedding.embedding import BaseEmbedding
from typing import Optional

class VectorStorageFactory:
    @staticmethod
    def create_vector_storage(
        knowledge_base: bool,
        use_bedrock_kb: bool,
        embedding: BaseEmbedding,
        opensearch_host: Optional[str] = None,
        opensearch_port: Optional[int] = None,
        opensearch_username: Optional[str] = None,
        opensearch_password: Optional[str] = None,
        index_id: Optional[str] = None,
        knowledge_base_id: Optional[str] = None,
        aws_region: str = "us-east-1"
    ) -> VectorStorage:
        """
        Factory method to return the appropriate vector storage client.

        :param use_bedrock_kb: Boolean flag to decide which storage to use.
        :param embedding: The embedding model to use.
        :param opensearch_host: OpenSearch host (Only needed for OpenSearch).
        :param opensearch_port: OpenSearch port (Only needed for OpenSearch).
        :param opensearch_username: OpenSearch username (Only needed for OpenSearch).
        :param opensearch_password: OpenSearch password (Only needed for OpenSearch).
        :param index_id: OpenSearch index ID (Only needed for OpenSearch).
        :param knowledge_base_id: Bedrock Knowledge Base ID (Only needed for Bedrock KB).
        :param aws_region: AWS region for Bedrock Knowledge Base (Defaults to "us-east-1").
        :return: An instance of `VectorStorage` (either `OpenSearchClient` or `BedrockKnowledgeBaseStorage`).
        """
        if not knowledge_base:
            return NoOpsVectorStorage()

        if use_bedrock_kb:
            if not knowledge_base_id:
                raise ValueError("Knowledge Base ID must be provided when using Bedrock Knowledge Base.")
            
            return BedrockKnowledgeBaseStorage(
                knowledge_base_id=knowledge_base_id,
                region=aws_region,
                embedder=embedding
            )
        
        if not (opensearch_host and opensearch_port and opensearch_username and opensearch_password and index_id):
            raise ValueError("All OpenSearch parameters must be provided when using OpenSearch.")

        return OpenSearchClient(
            opensearch_host, 
            opensearch_port,
            opensearch_username, 
            opensearch_password,
            index_id,
            embedder=embedding
        )



import json
from flotorch_core.storage.storage import StorageProvider
from typing import List

"""
This class is responsible for reading the JSON data from the storage.
"""
class JSONReader:
    """
    Initializes the JSONReader object.
    :param storage_provider: The storage provider.
    """

    def __init__(self, storage_provider: StorageProvider):
        self.storage_provider = storage_provider

    """
    Reads the JSON data from the storage.
    :param path: The path of the JSON file.
    :return: The JSON data.
    """
    def read(self, path:str) -> dict:
        data = "".join(chunk.decode("utf-8") for chunk in self.storage_provider.read(path))
        return json.loads(data)
    
    """
    Reads the JSON data from the storage and converts it to a model object.
    :param path: The path of the JSON file.
    :param model_class: The class of the model to convert the JSON data to.
    :return: The model object.
    """
    def read_as_model(self, path: str, model_class: type) -> List[object]:
        data = self.read(path)

        if isinstance(data, list):
            return [model_class(**item) for item in data]
        
        return [model_class(**data)]


from flotorch_core.storage.db.vector.vector_storage import VectorStorage
from flotorch_core.rerank.rerank import BedrockReranker

class RerankedVectorStorage(VectorStorage):

    def __init__(self, vectorStorage: VectorStorage, bedrockReranker: BedrockReranker):
        self.vectorStorage = vectorStorage
        self.reranker = bedrockReranker
    
    def search(self, query: str, knn: int, hierarchical=False):
        results = self.vectorStorage.search(query, knn, hierarchical)
        return self.reranker.rerank_documents(query, results)

    def embed_query(self, embedding, knn, hierarical=False):
        self.vectorStorage.embed_query(embedding, knn, hierarical)


import io

from PyPDF2 import PdfReader

from flotorch_core.storage.storage import StorageProvider


class PDFReader:
    def __init__(self, storage_provider: StorageProvider):
        self.storage_provider = storage_provider

    def read_pdf(self, path: str) -> list[str]:
        text = []
        for data in self.storage_provider.read(path):
            if data is not None:
                text.append(self._read_pdf(data))
        return text

    def _read_pdf(self, data: bytes) -> str:
        stream = io.BytesIO(data)
        reader = PdfReader(stream)
        text = ""
        for page in reader.pages:
            text += page.extract_text() or ""
        return text



from abc import ABC, abstractmethod
from typing import List, Dict, Any

"""
This class is responsible for storing the data in the database.
"""
class DBStorage(ABC):

    @abstractmethod
    def read(self, key) -> dict:
        pass

    @abstractmethod
    def write(self, item: dict):
        pass

    def bulk_write(self, items: List[dict]):
        for item in items:
            self.write(item)



from flotorch_core.logger.logger_provider import LoggerProvider

class Logger:
    """
    Main logger class that delegates logging to a logger provider.
    """

    _instance = None  # Singleton instance

    def __new__(cls, provider: LoggerProvider = None):
        if cls._instance is None:
            if provider is None:
                raise ValueError("LoggerProvider must be provided for the first initialization.")
            cls._instance = super(Logger, cls).__new__(cls)
            cls._instance.provider = provider
        return cls._instance

    def log(self, level: str, message: str) -> None:
        self.provider.log(level, message)

    def info(self, message: str) -> None:
        self.log("INFO", message)

    def error(self, message: str) -> None:
        self.log("ERROR", message)

    def warning(self, message: str) -> None:
        self.log("WARNING", message)

    def debug(self, message: str) -> None:
        self.log("DEBUG", message)



from abc import ABC, abstractmethod
from typing import Any

class ConfigProvider(ABC):
    """
    Abstract base class for configuration providers.
    """
    @abstractmethod
    def get(self, key: str, default: Any = None) -> Any:
        """
        Retrieves the configuration value for the given key.
        Args:
            key (str): The configuration key to retrieve.
            default (str): The default value when key not present
        Returns:
            Any: The value of the configuration key.
        """
        pass



from typing import List

from langchain.text_splitter import CharacterTextSplitter

from flotorch_core.chunking.chunking import Chunk
from flotorch_core.chunking.fixedsize_chunking import FixedSizeChunker


class HieraricalChunker(FixedSizeChunker):
    def __init__(self, chunk_size: int, chunk_overlap: int, parent_chunk_size: int):
        super().__init__(chunk_size, chunk_overlap)
        self.parent_chunk_size = self.tokens_per_charecter * parent_chunk_size
        if self.parent_chunk_size <= 0:
            raise ValueError("parent_chunk_size must be positive")
        if self.chunk_size > self.parent_chunk_size:
            raise ValueError("child_chunk_size must be less than parent chunking size")

    def chunk(self, data: str) -> List[Chunk]:
        if not data:
            raise ValueError("Input text cannot be empty or None")

        data = self._clean_data(data)
        self.parent_text_splitter = CharacterTextSplitter(
            separator=self.space,
            chunk_size=self.parent_chunk_size,
            chunk_overlap=0,  # Can change this at a later point of time
            length_function=len,
            is_separator_regex=False
        )
        self.child_text_splitter = CharacterTextSplitter(
            separator=self.space,
            chunk_size=self.chunk_size,
            chunk_overlap=self.chunk_overlap,
            length_function=len,
            is_separator_regex=False
        )
        parent_chunks = self.parent_text_splitter.split_text(data)
        overall_chunks = []
        for parent_chunk in parent_chunks:
            chunk_object = Chunk(parent_chunk)
            child_chunks = self.child_text_splitter.split_text(parent_chunk)
            for child_chunk in child_chunks:
                child_chunk_object = Chunk(child_chunk)
                chunk_object.add_child(child_chunk_object)

            overall_chunks.append(chunk_object)
        return overall_chunks



import boto3
from flotorch_core.storage.db.db_storage import DBStorage
from botocore.exceptions import ClientError
from typing import List, Dict, Any

class DynamoDB(DBStorage):
    def __init__(self, table_name, region_name='us-east-1'):
        self.table_name = table_name
        self.dynamodb = boto3.resource('dynamodb', region_name=region_name)
        self.table = self.dynamodb.Table(table_name)

    def write(self, item: dict):
        try:
            self.table.put_item(Item=item)
            return True
        except ClientError as e:
            print(f"Error writing to DynamoDB: {e}")
            return False

    def read(self, key) -> dict:
        try:
            response = self.table.get_item(Key=key)
            return response.get('Item', None)
        except ClientError as e:
            print(f"Error reading from DynamoDB: {e}")
            return None
    
    def bulk_write(self, items: list):
        with self.table.batch_writer() as batch:
            for item in items:
                batch.put_item(Item=item)
        return True
    
    def update(self, key: Dict[str, Any], data: Dict[str, Any]) -> bool:
        """
        Update method accepts:
        - `key`: Unique identifier to find the record (e.g., {'id': 123})
        - `data`: Fields to be updated with new values (e.g., {'status': 'completed'})
        """
        try:
            # Dynamically construct UpdateExpression and ExpressionAttributeValues
            update_expression = "SET " + ", ".join(f"{k} = :{k}" for k in data.keys())
            expression_values = {f":{k}": v for k, v in data.items()}

            self.table.update_item(
                Key=key,
                UpdateExpression=update_expression,
                ExpressionAttributeValues=expression_values,
                ReturnValues="UPDATED_NEW"
            )
            return True
        except ClientError as e:
            print(f"Error updating DynamoDB: {e}")
            return False
    


import unittest

from chunking.chunking import Chunk
from embedding.cohere_embedding import CohereEmbedding
from embedding.titanv1_embedding import TitanV1Embedding
from embedding.titanv2_embedding import TitanV2Embedding


class TestCohereEmbed(unittest.TestCase):
    def setUp(self):
        self.embedder = CohereEmbedding(model_id="test_model", region="us-west-2")

    def test_prepare_chunk(self):
        chunk = Chunk(data="test data")
        payload = self.embedder._prepare_chunk(chunk)
        self.assertEqual(payload, {"texts": ["test data"], "input_type": "search_document"})


class TestTitanV1Embed(unittest.TestCase):
    def setUp(self):
        self.embedder = TitanV1Embedding(model_id="test_model", region="us-west-2")

    def test_prepare_chunk(self):
        chunk = Chunk(data="test data")
        payload = self.embedder._prepare_chunk(chunk)
        self.assertEqual(payload, {"inputText": "test data", "embeddingConfig": {"outputEmbeddingLength": 256}})

    def test_extract_embedding(self):
        response = {"embedding": [0.1, 0.2, 0.3]}
        embedding = self.embedder.extract_embedding(response)
        self.assertEqual(embedding, [0.1, 0.2, 0.3])


class TestTitanV2Embed(unittest.TestCase):
    def setUp(self):
        self.embedder = TitanV2Embedding(model_id="test_model", region="us-west-2")

    def test_prepare_chunk(self):
        chunk = Chunk(data="test data")
        payload = self.embedder._prepare_chunk(chunk)
        self.assertEqual(payload, {"inputText": "test data", "dimensions": 256, "normalize": True})

    def test_extract_embedding(self):
        response = {"embedding": [0.1, 0.2, 0.3]}
        embedding = self.embedder.extract_embedding(response)
        self.assertEqual(embedding, [0.1, 0.2, 0.3])


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod
import logging

class LoggerProvider(ABC):
    """
    Abstract base class for logger providers.
    """

    @abstractmethod
    def log(self, level: str, message: str) -> None:
        """
        Logs a message at the given level.
        Args:
            level (str): The logging level (e.g., 'INFO', 'ERROR').
            message (str): The message to log.
        """
        pass

    @abstractmethod
    def get_logger(self) -> logging.Logger:
        """
        Returns the underlying logger instance.
        """
        pass



import json
import boto3
from botocore.exceptions import ClientError
import sagemaker
from sagemaker.jumpstart.model import JumpStartModel
from sagemaker.huggingface import HuggingFaceModel, get_huggingface_llm_image_uri
from sagemaker.session import Session

from flotorch_core.logger.global_logger import get_logger


logger = get_logger()

EMBEDDING_MODELS = {
    "huggingface-sentencesimilarity-bge-large-en-v1-5": {
        "model_name": "bge-large",
        "model_source": "jumpstart",
        "dimension": 1024,
        "instance_type": "ml.g5.2xlarge",
        "input_key": "text_inputs"
    },
    "huggingface-sentencesimilarity-bge-m3": {
        "model_name": "bge-m3",
        "model_source": "jumpstart",
        "dimension": 1024,
        "instance_type": "ml.g5.2xlarge",
        "input_key": "text_inputs"
    },
    "huggingface-textembedding-gte-qwen2-7b-instruct": {
        "model_name": "qwen",
        "model_source": "jumpstart",
        "dimension": 3584,
        "instance_type": "ml.g5.2xlarge",
        "input_key": "inputs"
    }
}


INFERENCER_MODELS = {
    "meta-textgeneration-llama-3-1-8b-instruct": {
        "model_source": "jumpstart",
        "instance_type": "ml.g5.2xlarge"
    },
    "huggingface-llm-falcon-7b-instruct-bf16": {
        "model_source": "jumpstart",
        "instance_type": "ml.g5.2xlarge"
    },
    "meta-textgeneration-llama-3-3-70b-instruct": {
        "model_source": "jumpstart",
        "instance_type": "ml.p4d.24xlarge"
    }
    ,
    "deepseek-ai/DeepSeek-R1-Distill-Llama-8B": {
        "model_source": "huggingface",
        "instance_type": "ml.g5.2xlarge"
    }
    ,
    "deepseek-ai/DeepSeek-R1-Distill-Qwen-1.5B": {
        "model_source": "huggingface",
        "instance_type": "ml.g5.xlarge"
    }
    ,
    "deepseek-ai/DeepSeek-R1-Distill-Qwen-7B": {
        "model_source": "huggingface",
        "instance_type": "ml.g5.xlarge"
    }
    ,
    "deepseek-ai/DeepSeek-R1-Distill-Qwen-14B": {
        "model_source": "huggingface",
        "instance_type": "ml.g6e.12xlarge"
    }
}


class SageMakerUtils:
    @staticmethod
    def check_endpoint_exists(sagemaker_client, endpoint_name):        
        try:
            response = sagemaker_client.describe_endpoint(EndpointName=endpoint_name)
            status = response['EndpointStatus']
            print(f"Endpoint '{endpoint_name}' status: {status}")
            return status == 'InService' or status == 'Creating'  # Returns True if InService, False otherwise

        except ClientError as e:
            error_code = e.response['Error']['Code']
            
            if error_code == 'ValidationException' and 'Could not find endpoint' in e.response['Error']['Message']:
                print(f"Endpoint '{endpoint_name}' does not exist.")
                return False  # Return False if the endpoint does not exist
            
            else:
                # Re-raise unexpected exceptions
                print(f"An unexpected error occurred: {e}")
                raise

    @staticmethod
    def create_jumpstart_endpoint(sagemaker_client, instance_type, region, role, model_id: str, endpoint_name: str) -> bool:
        """
        Creates a SageMaker endpoint for JumpStart models.
        Reuses existing model and endpoint configuration if available.

        Args:
            model_id (str): The model ID for the SageMaker JumpStart model.
            endpoint_name (str): The name of the SageMaker endpoint to be created.

        Returns:
            bool: True if the endpoint is successfully created, False otherwise.
        """
        try:
            boto_session = boto3.Session(region_name=region)
            sagemaker_session = sagemaker.Session(boto_session=boto_session)

            # Initialize JumpStart model
            model = JumpStartModel(
                role=role,
                model_id=model_id,
                sagemaker_session=sagemaker_session
            )

            # Check if the endpoint configuration exists
            try:
                sagemaker_client.describe_endpoint_config(
                    EndpointConfigName=endpoint_name
                )
                logger.info(f"Endpoint configuration '{endpoint_name}' exists. Deploying endpoint.")

                # Deploy the endpoint using the existing configuration
                sagemaker_client.create_endpoint(
                    EndpointName=endpoint_name,
                    EndpointConfigName=endpoint_name
                )
            except sagemaker_client.exceptions.ClientError as e:
                error_code = e.response['Error']['Code']
                if error_code == 'ValidationException' and 'Could not find endpoint configuration' in e.response['Error']['Message']:
                    logger.info(f"Endpoint configuration '{endpoint_name}' does not exist. Deploying using model.deploy().")

                    # Use model.deploy to handle everything
                    model.deploy(
                        initial_instance_count=1,
                        instance_type=instance_type,
                        endpoint_name=endpoint_name,
                        accept_eula=True
                    )
                else:
                    logger.error(f"Error while checking endpoint configuration: {e}")
                    raise

            logger.info(f"Successfully created endpoint '{endpoint_name}' for model '{model_id}'.")
            return True

        except sagemaker_client.exceptions.ResourceLimitExceeded as e:
            logger.error(f"Resource limit exceeded while creating endpoint: {e}")
            return False

        except Exception as e:
            logger.error(f"Error while creating endpoint '{endpoint_name}' for model '{model_id}': {e}")
            return False


    @staticmethod
    def create_huggingface_endpoint(sagemaker_client, instance_type, model_id: str, endpoint_name: str, role: str, region_name: str) -> bool:
        """
        Creates a SageMaker endpoint for HuggingFace models.
        - If the endpoint config exists, it deploys using the existing configuration.
        - If not, it does a full model deployment.

        Args:
            sagemaker_client (boto3.client): Boto3 SageMaker client.
            instance_type (str): AWS instance type for deployment.
            model_id (str): Hugging Face model ID.
            endpoint_name (str): Name of the SageMaker endpoint.
            role (str): IAM role for SageMaker execution.
            region_name (str): AWS region.

        Returns:
            bool: True if the endpoint is successfully created, False otherwise.
        """
        try:
            session = Session(boto_session=boto3.Session(region_name=region_name))
            hub = {
                'HF_MODEL_ID': model_id,
                'SM_NUM_GPUS': json.dumps(1)
            }
            huggingface_model = HuggingFaceModel(
                image_uri=get_huggingface_llm_image_uri("huggingface", version="2.3.1", region=region_name),
                env=hub,
                role=role,
                sagemaker_session = session 
            )

            predictor = huggingface_model.deploy(
                initial_instance_count=1,
                instance_type=instance_type,
                endpoint_name=endpoint_name,
                container_startup_health_check_timeout=300,
            )
            
            return predictor
        
        except sagemaker_client.exceptions.ResourceLimitExceeded as e:
            logger.error(f"Resource limit exceeded while creating endpoint: {e}")
            return False

        except Exception as e:
            logger.error(f"Error while creating endpoint '{endpoint_name}' for model '{model_id}': {e}")
            return False

        
        
    @staticmethod
    def wait_for_endpoint_creation(sagemaker_client, endpoint_name: str, wait_interval: int = 5, timeout: int = 100000) -> bool:
        """
        Waits until the SageMaker endpoint is created successfully.

        Args:
            endpoint_name (str): The name of the SageMaker endpoint.
            wait_interval (int): Time (in seconds) to wait between status checks. Default is 30 seconds.
            timeout (int): Maximum time (in seconds) to wait for the endpoint creation. Default is 1800 seconds (30 minutes).

        Returns:
            bool: True if the endpoint is successfully created and in service, False if timed out or failed.
        """
        import time

        start_time = time.time()
        logger.info(f"Waiting for endpoint '{endpoint_name}' to be in service...")

        try:
            while time.time() - start_time < timeout:
                response = sagemaker_client.describe_endpoint(EndpointName=endpoint_name)
                status = response["EndpointStatus"]

                logger.info(f"Endpoint '{endpoint_name}' status: {status}")

                if status == "InService":
                    logger.info(f"Endpoint '{endpoint_name}' is now in service.")
                    return True
                elif status == "Failed":
                    logger.error(f"Endpoint '{endpoint_name}' creation failed.")
                    return False

                logger.error(f"waiting for endpoint creating'{endpoint_name}' , status: {status}")
                time.sleep(wait_interval)  # Wait before checking again

            logger.error(f"Timeout while waiting for endpoint '{endpoint_name}' to be created.")
            return False

        except sagemaker_client.exceptions.ResourceNotFound:
            logger.error(f"Endpoint '{endpoint_name}' not found. Ensure the creation process has started.")
            return False

        except Exception as e:
            logger.error(f"Error while checking endpoint status for '{endpoint_name}': {e}")
            return False
    
    @staticmethod
    def sanitize_name( name: str) -> str:
        """Sanitize the endpoint name to follow AWS naming conventions"""
        # Replace any character that's not alphanumeric or hyphen with hyphen
        import re
        name = re.sub(r'[^a-zA-Z0-9-]', '-', name)
        # Ensure it starts with a letter
        if not name[0].isalpha(): 
            name = 'n' + name
        # Truncate to 63 characters (AWS limit)
        return name[:63]


import unittest
from testcontainers.opensearch import OpenSearchContainer

from storage.db.vector.open_search import OpenSearchClient  # Assuming this is the class you want to test

class TestOpenSearchClient(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.opensearch_container = OpenSearchContainer(security_enabled=True)
        cls.opensearch_container.start()
        local_client = cls.opensearch_container.get_client()
        local_client.index(index="test", body={"test": "test"})
        cls.opensearch_client = OpenSearchClient(
            host=cls.opensearch_container.get_container_host_ip(),
                                        port=cls.opensearch_container.get_exposed_port(9200),
                                        username='admin',
                                        password='admin',
                                        index='test',
                                        use_ssl=True,
                                        verify_certs=False
                                        )
        
    @classmethod
    def tearDownClass(cls):
        cls.opensearch_container.stop()


    def test_write(self):
        body = {
            "name": "John Doe",
            "age": 30
        }
        # create open search client object at the class level and use it in the test
        self.opensearch_client.write(body)
    
    def test_read(self):
        body = {
            "query": {
                "match_all": {}
            }
        }
        # create open search client object at the class level and use it in the test
        response = self.opensearch_client.search(body)
        print(response)

if __name__ == '__main__':
    unittest.main()


from flotorch_core.inferencer.inferencer import BaseInferencer
from typing import List, Dict, Any, Tuple
from flotorch_core.logger.global_logger import get_logger
import boto3
import random

from flotorch_core.utils.bedrock_retry_handler import BedRockRetryHander


logger = get_logger()

class BedrockInferencer(BaseInferencer):
    """
    Bedrock-specific implementation of the BaseInferencer.
    """

    def __init__(self, model_id: str, region: str = "us-east-1", n_shot_prompts: int = 0, temperature: float = 0.7, n_shot_prompt_guide_obj: Dict[str, List[Dict[str, str]]] = None):
        """
        Initialize the BedrockInferencer with Bedrock-specific parameters.

        Args:
            model_id (str): Identifier for the Bedrock model.
            region (str): AWS region where the Bedrock service is deployed.
            n_shot_prompts (int): Number of examples to include in few-shot learning.
            temperature (float): Sampling temperature for response generation.
            n_shot_prompt_guide_obj (Dict[str, List[Dict[str, str]]]): Guide object for few-shot examples.
        """
        super().__init__(model_id, region, n_shot_prompts, temperature, n_shot_prompt_guide_obj)
        self.client = boto3.client(
            service_name='bedrock-runtime',
            region_name=region
        )

    @BedRockRetryHander()
    def generate_text(self, user_query: str, context: List[Dict] = None) -> Tuple[Dict[Any, Any], str]:
        """
        Generate a response based on the user query and context using Bedrock.
        """
        try:
            system_prompt, messages = self.generate_prompt(user_query, context)
            
            inference_config = {
                "maxTokens": 512, 
                "temperature": self.temperature, 
                "topP": 0.9
            }
            
            skip_system_param = self.model_id in ("amazon.titan-text-express-v1", "amazon.titan-text-lite-v1", "mistral.mistral-7b-instruct-v0:2")
            request_params = {
                "modelId": self.model_id,
                "messages": ([self._prepare_conversation(role="user", message=system_prompt)] if skip_system_param else []) + messages,
                "inferenceConfig": inference_config
            }
            
            if not skip_system_param:
                request_params["system"] = [{"text": system_prompt}]
            
            response = self.client.converse(**request_params)
            
            metadata = {}
            if 'usage' in response:
                for key, value in response['usage'].items():
                    metadata[key] = value
            if 'metrics' in response:
                for key, value in response['metrics'].items():
                    metadata[key] = value
            
            return metadata, self._extract_response(response)
        except Exception as e:
            logger.error(f"Error generating text with Bedrock: {str(e)}")
            raise

    def generate_prompt(self, user_query: str, context: List[Dict] = None) -> Tuple[str, List[Dict[str, Any]]]:
        """
        Construct a prompt for the Bedrock inferencer based on the user query and context.
        """
        messages = []
        context_text = ""
        
        # Validate n_shot_prompt
        if self.n_shot_prompts < 0:
            raise ValueError("n_shot_prompts must be non-negative")
        
        default_prompt = "You are a helpful assistant. Use the provided context to answer questions accurately. If you cannot find the answer in the context, say so"
        # Get system prompt
        system_prompt = default_prompt if not self.n_shot_prompt_guide_obj or not self.n_shot_prompt_guide_obj.get("system_prompt") else self.n_shot_prompt_guide_obj.get("system_prompt")
        
        # Process context
        if context:
            context_text = self.format_context(context)
            if context_text:
                messages.append(self._prepare_conversation(role="user", message=context_text))
        
        base_prompt = self.n_shot_prompt_guide_obj.get("user_prompt", "") if self.n_shot_prompt_guide_obj else ""
        if base_prompt:
            messages.append(self._prepare_conversation(role="user", message=base_prompt))
        
        # Get examples
        examples = self.n_shot_prompt_guide_obj.get("examples", []) if self.n_shot_prompt_guide_obj else []
        selected_examples = random.sample(examples, self.n_shot_prompts) if len(examples) > self.n_shot_prompts else examples
        
        # Format examples
        for example in selected_examples:
            if 'example' in example:
                messages.append(self._prepare_conversation(role="user", message=example['example']))
            elif 'question' in example and 'answer' in example:
                messages.append(self._prepare_conversation(role="user", message=example['question']))
                messages.append(self._prepare_conversation(role="assistant", message=example['answer']))
        
        logger.info(f"Using {self.n_shot_prompts} shot prompt with {len(selected_examples)} examples")
        
        # Add user query
        messages.append(self._prepare_conversation(role="user", message=user_query))
        
        return system_prompt, messages

    def _prepare_conversation(self, message: str, role: str) -> Dict[str, Any]:
        """Formats a message and role into a conversation dictionary."""
        if not message or not role:
            logger.error("Error in parsing message or role")
        
        return {"role": role, "content": [{"text": message}]}

    def format_context(self, context: List[Dict[str, str]]) -> str:
        """Format context documents into a single string."""
        if not context or len(context) == 0:
            return ""
        
        context_text = "\n".join([
            f"Context {i+1}:\n{doc.get('text', '')}"
            for i, doc in enumerate(context)
        ])
        logger.debug(f"Formatted context text length: {len(context_text)}")
        return context_text

    def _extract_response(self, response: Dict) -> str:
        """
        Extract the generated text from the Bedrock response.

        Args:
            response (Dict): Response from the Bedrock API.

        Returns:
            str: Extracted text from the response.
        """
        response_text = response["output"]["message"]["content"][0]["text"]
        logger.debug(f"Response length: {len(response_text)}")
        return response_text



import logging
import os
from typing import Generator

from .storage import StorageProvider

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class LocalStorageProvider(StorageProvider):
    """
    Local storage provider
    """

    def __init__(self):
        """
        Initializes the LocalStorageProvider class.
        """
        super().__init__()

    def get_path(self, uri: str) -> str:
        return uri.split("://")[1]

    def write(self, path: str, data: bytes) -> None:
        """
        Writes data to the specified path in local storage.
        Args:
            path (str): The path to write the data to in local storage.
            data (bytes): The data to write to local storage.
        """
        logger.info(f'Writing data to local storage: {data}')
        if os.path.isdir(path):
            path = os.path.join(path, 'tmp.data')
        with open(path, 'wb') as file:
            file.write(data)

    def read(self, path) -> Generator[bytes, None, None]:
        """
        Reads data from the specified path in local storage.
        Args:
            path (str): The path to read the data from in local storage.
        Returns:
            Generator[bytes, None, None]: A generator that yields the data read from local storage.
        """
        logger.info('Reading data from local storage')
        if os.path.isdir(path):
            yield from self._read_directory(path)
        else:
            with open(path, "rb") as file:
                yield file.read()

    def _read_directory(self, path):
        """
        Reads all files in a directory.
        Args:
            path (str): The path to the directory to read.
        Returns:
            Generator[bytes, None, None]: A generator that yields the data read from each file in the directory.
        """
        for filename in os.listdir(path):
            file_path = os.path.join(path, filename)
            if os.path.isfile(file_path):
                with open(path, "rb") as file:
                    return file.read()



from urllib.parse import urlparse
from flotorch_core.storage.storage import StorageProvider
from flotorch_core.storage.s3_storage import S3StorageProvider
from flotorch_core.storage.local_storage import LocalStorageProvider

class StorageProviderFactory:
    """
    Factory to create storage providers based on the URI scheme.
    """
    @staticmethod
    def create_storage_provider(uri: str) -> StorageProvider:
        parsed = urlparse(uri)
        if parsed.scheme == "s3":
            bucket = parsed.netloc
            return S3StorageProvider(bucket)
        elif parsed.scheme == "" or parsed.scheme == "file":
            return LocalStorageProvider()
        else:
            raise ValueError(f"Unsupported storage scheme: {parsed.scheme}")



import os
from typing import Any
from .config_provider import ConfigProvider

class EnvConfigProvider(ConfigProvider):
    """
    Configuration provider that fetches values from environment variables.
    """

    def get(self, key: str, default: Any = None) -> Any:
        """
        Retrieves the value of the environment variable or .env file.
        Args:
            key (str): The environment variable key.
            default (Any): The default value if the key is not found.
        Returns:
            Any: The value of the environment variable or the default.
        """
        return os.getenv(key, default)



import json
from typing import List, Dict, Any
from abc import abstractmethod
import boto3

from flotorch_core.chunking.chunking import Chunk
from flotorch_core.utils.bedrock_retry_handler import BedRockRetryHander
from .embedding import BaseEmbedding, Embeddings, EmbeddingMetadata


class BedRockEmbedding(BaseEmbedding):
    def __init__(self, model_id: str, region: str, dimensions: int = 256, normalize: bool = True) -> None:
        super().__init__(model_id, region, dimensions, normalize)
        self._application_json = "application/json"
        self.client = boto3.client("bedrock-runtime", region_name=self.region)

    @BedRockRetryHander()
    def embed(self, chunk: Chunk) -> Embeddings:
        payload = self._prepare_chunk(chunk)
        response = self._invoke_model(payload)
        metadata = self._extract_metadata(response)
        model_response = self._parse_model_response(response)
        return Embeddings(embeddings=self.extract_embedding(model_response),
                          metadata=metadata, text=chunk.data)

    def _invoke_model(self, payload: Dict[str, Any]) -> Dict[str, Any]:
        return self.client.invoke_model(
            modelId=self.model_id,
            contentType=self._application_json,
            accept=self._application_json,
            body=json.dumps(payload)
        )

    def _extract_metadata(self, response: Dict[str, Any]) -> EmbeddingMetadata:
        if not response or 'ResponseMetadata' not in response:
            return EmbeddingMetadata(input_tokens=0, latency_ms=0)

        headers = response['ResponseMetadata'].get('HTTPHeaders', {})
        return EmbeddingMetadata(
            input_tokens=headers.get('x-amzn-bedrock-input-token-count', 0),
            latency_ms=headers.get('x-amzn-bedrock-invocation-latency', 0)
        )

    def _parse_model_response(self, response: Dict[str, Any]) -> Dict[str, Any]:
        if 'body' not in response:
            raise ValueError("Invalid response format: 'body' not found.")
        return json.loads(response['body'].read())

    def extract_embedding(self, response: Dict[str, Any]) -> List[float]:
        return response["embeddings"]



from abc import abstractmethod
import boto3
from typing import Any, Dict, List
from flotorch_core.chunking.chunking import Chunk
from flotorch_core.embedding.embedding import EmbeddingMetadata, Embeddings
from flotorch_core.embedding.embedding import BaseEmbedding
from sagemaker.session import Session
from sagemaker.predictor import Predictor
from sagemaker.serializers import JSONSerializer
from sagemaker.deserializers import JSONDeserializer
from sagemaker.jumpstart.model import JumpStartModel
from sagemaker.huggingface import HuggingFaceModel, get_huggingface_llm_image_uri
import sagemaker
import numpy as np
import json
import time
from botocore.exceptions import ClientError

from flotorch_core.logger.global_logger import get_logger
from flotorch_core.utils.sagemaker_utils import SageMakerUtils, EMBEDDING_MODELS

logger = get_logger()

class SageMakerEmbedder(BaseEmbedding):
    def __init__(self, model_id: str, region: str, role_arn: str, dimensions: int = 256, normalize: bool = True) -> None:
        """
        Initializes the SageMakerEmbedder with the given model ID, region, and role ARN.
        Sets up necessary SageMaker runtime clients, session, and endpoint predictor.

        Args:
            model_id (str): The unique identifier for the model.
            region (str): The AWS region where the SageMaker services are hosted.
            role_arn (str): The ARN of the IAM role. Currently not used but included for future extensions.
        """

        # Initialize the base class
        super().__init__(model_id, region, dimensions, normalize)
        
        self.role = role_arn
        
        # Initialize the SageMaker runtime and client for general operations
        self.client = boto3.client("sagemaker-runtime", region_name=region)
        self.sagemaker_client = boto3.client('sagemaker', region_name=region)
        
        # Create a new SageMaker session
        self.session = Session(boto_session=boto3.Session(region_name=region))
        
        # Initialize additional embedding-related attributes
        self.embedding_model_id = model_id
        # self.embedding_model_endpoint_name = 'flotorch-embedding-endpoint'
        self.embedding_model_endpoint_name = f"{SageMakerUtils.sanitize_name(model_id)[:44]}-embedding-endpoint"
        
        self.embedding_dimension = EMBEDDING_MODELS.get(model_id, {}).get('dimension', 1024)
        
        self.wait_time = 5

        if not SageMakerUtils.check_endpoint_exists(self.sagemaker_client, self.embedding_model_endpoint_name):
            model_config = EMBEDDING_MODELS.get(model_id)
            SageMakerUtils.create_jumpstart_endpoint(self.sagemaker_client, model_config.get("instance_type"), self.region, self.role, model_id, self.embedding_model_endpoint_name)

        SageMakerUtils.wait_for_endpoint_creation(self.sagemaker_client, self.embedding_model_endpoint_name)
        
        # Ensure the endpoint exists or create it if necessary
        # self._ensure_endpoint_exists()
        
        # Initialize the predictor to interact with the SageMaker endpoint
        self.predictor = Predictor(
            endpoint_name=self.embedding_model_endpoint_name,
            sagemaker_session=self.session
        )
        
        # Set up the serializer and deserializer for the predictor
        self.predictor.serializer = JSONSerializer()
        self.predictor.deserializer = JSONDeserializer()

        self.embedding_predictor = self.predictor
        
        # Log initialization success
        logger.info(f"Initialized SageMakerEmbedder for model {model_id} in region {region}.")

    @abstractmethod
    def _prepare_chunk(self, chunk: Chunk) -> Dict:
        """
        Abstract method for preparing payload for SageMaker models.
        """
        pass
          
    def _check_model_status(self, endpoint_name, loop = True):
        """
        Check the status of the SageMaker endpoint and its configuration.
        
        This method performs the following:
        1. Checks if endpoint exists and is in service
        2. If endpoint is being created, waits until creation completes
        3. If no endpoint exists, checks for endpoint configuration
        4. If configuration exists, waits for endpoint creation to complete
        
        Args:
            endpoint_name (str): Name of the SageMaker endpoint to check
            
        Returns:
            str: 'InService' if endpoint is available and running
            
        Raises:
            Exception: If endpoint creation fails or has unexpected status
            ClientError: If neither endpoint nor configuration exists
        """
        try:
            while True: 
                # Poll endpoint status until it is in service or fails
                response = self.sagemaker_client.describe_endpoint(EndpointName=endpoint_name)
                
                if response['EndpointStatus'] == 'InService':
                    logger.info(f"Endpoint {endpoint_name} is in service.")
                    return 'InService'
                
                elif response['EndpointStatus'] == 'Failed':
                    logger.error(f"Endpoint {endpoint_name} creation failed.")
                    raise Exception(f"Endpoint {endpoint_name} creation failed.")
                
                elif response['EndpointStatus'] == 'Creating':
                    time.sleep(self.wait_time) # Pause before next status check
                    
                else:
                    raise Exception(f"Unexpected endpoint status: {response['EndpointStatus']}")
                
        except self.sagemaker_client.exceptions.ClientError as e: 
            # No endpoint exists - check if there's a configuration waiting to be deployed
            logger.info(f"Endpoint {endpoint_name} does not exist. Checking if endpoint configuration exists.")
            
            try: 
                # Look for endpoint configuration that may have been created by another process
                response = self.sagemaker_client.describe_endpoint_config(EndpointConfigName=endpoint_name)
                logger.info(f"Configuration for {endpoint_name} exists, waiting {self.wait_time} seconds for endpoint creation.")
                
                time.sleep(self.wait_time) # Allow time for endpoint creation to begin
                
                if loop:
                    _ = self._check_model_status(endpoint_name)  # Keep rechecking the endpoint status until it begins creation
                else:
                    raise
                
            except self.sagemaker_client.exceptions.ClientError: 
                # Neither endpoint nor configuration exists
                logger.info(f"Endpoint configuration does not exist.")
                raise
            
        except Exception as e:
            logger.error(f"Error checking endpoint status: {e}")
            raise

    def create_endpoint(self, endpoint_name: str, model_id: str) -> sagemaker.predictor.Predictor:
        """
        Creates a SageMaker endpoint for the specified model if it doesn't already exist.
        If the endpoint is already in service, returns the existing predictor.
        
        Args:
            endpoint_name (str): The name of the SageMaker endpoint to be created or fetched.
            model_id (str): The identifier for the model to be used in the endpoint.
        
        Returns:
            sagemaker.predictor.Predictor: A predictor object for the created or fetched endpoint.

        Raises:
            ValueError: If the provided model_id is not supported.
            ClientError: If there are AWS API errors during endpoint creation/access.
        """
        
        # Validate that model_id exists in our supported model configurations
        if model_id not in EMBEDDING_MODELS:
            raise ValueError(f"Unsupported model ID: {model_id}")

        # Create AWS and SageMaker sessions for API interactions
        boto_session = boto3.Session(region_name=self.region)
        sagemaker_session = sagemaker.Session(boto_session=boto_session)

        # Look up the appropriate instance type from model configurations
        instance_type = (EMBEDDING_MODELS.get(model_id))['instance_type']
        model_source = (EMBEDDING_MODELS.get(model_id))['model_source']
        
        try:
            # First check if a working endpoint already exists to avoid duplicate creation
            status = self._check_model_status(endpoint_name, False)
            if status == 'InService':
                # Endpoint exists and is healthy - create and return a predictor for it
                predictor = sagemaker.predictor.Predictor(
                    endpoint_name=endpoint_name,
                    sagemaker_session=sagemaker_session,
                    serializer=sagemaker.serializers.JSONSerializer(),
                    deserializer=sagemaker.deserializers.JSONDeserializer()
                )
                # Register the predictor with the appropriate model type handler
                self._assign_predictor(predictor, model_id)
                return predictor
                
        except self.sagemaker_client.exceptions.ClientError as e:
            # Handle case where endpoint doesn't exist yet
            if e.response['Error']['Code'] == 'ValidationException':
                try:
                    if model_source == "jumpstart":
                        # Initialize a new JumpStart model with the specified configuration
                        model = JumpStartModel(
                            role = self.role,
                            model_id=model_id,
                            sagemaker_session=sagemaker_session
                        )
                        
                        # Deploy the model to a new endpoint with the specified configuration
                        predictor = model.deploy(
                            initial_instance_count=1,
                            instance_type=instance_type,
                            endpoint_name=endpoint_name,
                            accept_eula=True  # Required for JumpStart models
                        )
                    # Check if the model source is huggingface    
                    elif model_source == "huggingface":
                        hub = {
                            'HF_MODEL_ID': model_id,
                            'SM_NUM_GPUS': json.dumps(1)
                        }
                        huggingface_model = HuggingFaceModel(
                            image_uri=get_huggingface_llm_image_uri("huggingface", version="2.3.1", region=self.region),
                            env=hub,
                            role=self.role, 
                        )

                        # deploy model to SageMaker Inference
                        predictor = huggingface_model.deploy(
                            initial_instance_count=1,
                            instance_type=instance_type,
                            endpoint_name=endpoint_name,
                            container_startup_health_check_timeout=300,
                        )
                    
                    # Register the new predictor with the appropriate model type handler
                    self._assign_predictor(predictor, model_id)
                    return predictor
                
                except self.sagemaker_client.exceptions.ClientError as e:
                    # Handle race condition where another process started creating the endpoint
                    # between our existence check and creation attempt
                    logger.info(f"Error creating endpoint: {e}")
                    if e.response['Error']['Code'] == 'ValidationException':
                        
                        logger.info(f"A new endpoint creation intercepted while attempting to create new endpoint, waiting.")
                        time.sleep(self.wait_time) # Allow the other process's endpoint to finish creating
                        
                        status = self._check_model_status(endpoint_name)
                        
                        if status == 'InService':
                            logger.info(f"Found the new endpoint, creating the predictor.")
                            # Create predictor for the endpoint that the other process created
                            predictor = sagemaker.predictor.Predictor(
                                endpoint_name=endpoint_name,
                                sagemaker_session=sagemaker_session,
                                serializer=sagemaker.serializers.JSONSerializer(),
                                deserializer=sagemaker.deserializers.JSONDeserializer()
                            )
                            
                            # Register with appropriate model type handler
                            self._assign_predictor(predictor, model_id)
                            
                            return predictor
                        
                    elif e.response['Error']['Code'] == 'ResourceLimitExceeded':
                        logger.error(f"Resource limit exceeded while creating endpoint: {e}")
                        raise
            
            # Re-raise any unexpected AWS API errors
            raise

    def _assign_predictor(self, predictor: sagemaker.predictor.Predictor, model_id: str):
        """
        Assigns the appropriate predictor based on the provided model_id. The predictor is assigned to either 
        the embedding or inferencing predictor attributes, depending on the model type.

        Args:
            predictor (sagemaker.predictor.Predictor): The SageMaker predictor to be assigned.
            model_id (str): The model ID which determines whether the predictor is for embedding or inferencing.
        
        """
        # Assign predictor for embedding models
        if model_id in EMBEDDING_MODELS:
            self.embedding_predictor = predictor
            self.embedding_dimension = EMBEDDING_MODELS[model_id]['dimension']
            self.embedding_model_id = model_id
            logger.info(f"Assigned embedding predictor for model: {model_id}")
        
        # Log an error if the model_id doesn't match any known type
        else:
            logger.error(f"Model ID {model_id} is not recognized as an embedding or inferencing model.")

    def _invoke_model(self, payload: Dict[str, Any]) -> Dict[str, Any]:
        return self.embedding_predictor.predict(payload)
    
    def _extract_metadata(self, chunk: Chunk, latency: int) -> EmbeddingMetadata:
        return EmbeddingMetadata(
            input_tokens = len(chunk.data) // 4,
            latency_ms = latency
        )
    
    def _parse_model_response(self, response: Dict[str, Any]) -> Dict[str, Any]:
        # If the response is in byte format, decode it
        if isinstance(response, (bytes, bytearray)):
            response = json.loads(response.decode('utf-8'))
        elif isinstance(response, str):
            response = json.loads(response)

        # Extract the embedding from the response
        if isinstance(response, dict) and 'embedding' in response:
            embedding = np.array(response['embedding'][0] if isinstance(response['embedding'], list) else response['embedding'])
        else:
            embedding = np.array(response[0] if isinstance(response, list) else response)

        embedding = embedding.flatten()

        # Normalize the embedding to unit length
        embedding = embedding / np.linalg.norm(embedding)

        # Check if the embedding dimension matches the expected value (1024)
        if len(embedding) != self.embedding_dimension:
            logger.warning(f"Embedding dimension mismatch. Expected 1024, got {len(embedding)}")
            # Adjust the dimension by truncating or padding
            if len(embedding) > self.embedding_dimension:
                embedding = embedding[:self.embedding_dimension]
            else:
                embedding = np.pad(embedding, (0, self.embedding_dimension - len(embedding)))

        return embedding.tolist()

    def embed(self, chunk: Chunk) -> Embeddings:
        if not self.predictor:
            raise ValueError("Embedding predictor not initialized")

        if not chunk.data or not chunk.data.strip():
            raise ValueError("Input text cannot be empty")

        try:
            payload = self._prepare_chunk(chunk)
            start_time = time.time()
            response = self._invoke_model(payload)
            latency = int((time.time() - start_time) * 1000)
            metadata = self._extract_metadata(chunk, latency)
            model_response = self._parse_model_response(response)
            return Embeddings(embeddings=model_response, metadata=metadata, text=chunk.data)
        except Exception as e:
            # Log detailed error information for debugging
            logger.error("Error in get_embedding: %s", str(e))
            logger.error("Model ID: %s", self.embedding_model_id)
            logger.error("Input text length: %d", len(chunk.data))
            logger.error("Input data: %s", json.dumps(payload, indent=2))

            if 'response' in locals():
                logger.error("Response structure: %s", type(response))
                try:
                    logger.error("Response content: %s", json.dumps(response, indent=2))
                except Exception as json_error:
                    logger.error("Response content (raw): %s", response)
            raise
    


class EmbeddingRegistry:
    def __init__(self):
        self._models = {}

    def register_model(self, model_id, embedding_class):
        self._models[model_id] = embedding_class

    def get_model(self, model_id):
        embedding_class = self._models.get(model_id)
        if not embedding_class:
            raise ValueError(f"Model '{model_id}' not found in the registry.")
        return embedding_class

# Global registry instance
embedding_registry = EmbeddingRegistry()

def register(model_id):
    def decorator(cls):
        embedding_registry.register_model(model_id, cls)
        return cls
    return decorator


from flotorch_core.logger.logger import Logger
from flotorch_core.logger.console_logger_provider import ConsoleLoggerProvider

def get_logger():
    """
    Returns a singleton logger instance.
    """
    provider = ConsoleLoggerProvider()
    return Logger(provider)





from typing import Dict, List
from flotorch_core.inferencer.bedrock_inferencer import BedrockInferencer
from flotorch_core.inferencer.inferencer import BaseInferencer
from flotorch_core.inferencer.sagemaker_inferencer import SageMakerInferencer


class InferencerProviderFactory:
    """
    Factory to create inferencer based on the service name.
    """
    @staticmethod
    def create_inferencer_provider(service: str, model_id: str, region: str, arn_role: str, n_shot_prompts: int = 0, temperature: float = 0.7, n_shot_prompt_guide_obj: Dict[str, List[Dict[str, str]]] = None) -> BaseInferencer:
        if service == "bedrock":
            return BedrockInferencer(model_id, region, n_shot_prompts, temperature, n_shot_prompt_guide_obj)
        elif service == "sagemaker":
            return SageMakerInferencer(model_id, region, arn_role, n_shot_prompts, temperature, n_shot_prompt_guide_obj)
        else:
            raise ValueError(f"Unsupported service scheme: {service}")
